@@ -22,14 +22,19 @@ import warnings
 from datetime import datetime
 from decimal import Decimal
 from enum import Enum
-from typing import List, Optional, Tuple, Union
-from urllib.parse import ParseResult, urlparse
-from uuid import uuid4
+from typing import Any, Iterable, Optional, Tuple, Union
+
+from sortedcontainers import SortedSet
 
-from . import OrganizationalContact, OrganizationalEntity, Tool, XsUri
-from .impact_analysis import ImpactAnalysisAffectedStatus, ImpactAnalysisJustification, ImpactAnalysisResponse, \
-    ImpactAnalysisState
 from ..exception.model import MutuallyExclusivePropertiesException, NoPropertiesProvidedException
+from . import ComparableTuple, OrganizationalContact, OrganizationalEntity, Tool, XsUri
+from .bom_ref import BomRef
+from .impact_analysis import (
+    ImpactAnalysisAffectedStatus,
+    ImpactAnalysisJustification,
+    ImpactAnalysisResponse,
+    ImpactAnalysisState,
+)
 
 """
 This set of classes represents the data that is possible about known Vulnerabilities.
@@ -54,7 +59,7 @@ class BomTargetVersionRange:
         See the CycloneDX schema: https://cyclonedx.org/docs/1.4/#type_vulnerabilityType
     """
 
-    def __init__(self, version: Optional[str] = None, version_range: Optional[str] = None,
+    def __init__(self, *, version: Optional[str] = None, version_range: Optional[str] = None,
                  status: Optional[ImpactAnalysisAffectedStatus] = None) -> None:
         if not version and not version_range:
             raise NoPropertiesProvidedException(
@@ -106,6 +111,23 @@ class BomTargetVersionRange:
     def status(self, status: Optional[ImpactAnalysisAffectedStatus]) -> None:
         self._status = status
 
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, BomTargetVersionRange):
+            return hash(other) == hash(self)
+        return False
+
+    def __lt__(self, other: Any) -> bool:
+        if isinstance(other, BomTargetVersionRange):
+            return ComparableTuple((self.version, self.range, self.status)) < \
+                ComparableTuple((other.version, other.range, other.status))
+        return NotImplemented
+
+    def __hash__(self) -> int:
+        return hash((self.version, self.range, self.status))
+
+    def __repr__(self) -> str:
+        return f'<BomTargetVersionRange version={self.version}, version_range={self.range}, status={self.status}>'
+
 
 class BomTarget:
     """
@@ -120,9 +142,9 @@ class BomTarget:
         See the CycloneDX schema: https://cyclonedx.org/docs/1.4/#type_vulnerabilityType
     """
 
-    def __init__(self, ref: str, versions: Optional[List[BomTargetVersionRange]] = None) -> None:
+    def __init__(self, *, ref: str, versions: Optional[Iterable[BomTargetVersionRange]] = None) -> None:
         self.ref = ref
-        self.versions = versions
+        self.versions = SortedSet(versions or [])
 
     @property
     def ref(self) -> str:
@@ -136,15 +158,34 @@ class BomTarget:
         self._ref = ref
 
     @property
-    def versions(self) -> Optional[List[BomTargetVersionRange]]:
+    def versions(self) -> "SortedSet[BomTargetVersionRange]":
         """
         Zero or more individual versions or range of versions.
+
+        Returns:
+            Set of `BomTargetVersionRange`
         """
         return self._versions
 
     @versions.setter
-    def versions(self, versions: Optional[List[BomTargetVersionRange]]) -> None:
-        self._versions = versions
+    def versions(self, versions: Iterable[BomTargetVersionRange]) -> None:
+        self._versions = SortedSet(versions)
+
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, BomTarget):
+            return hash(other) == hash(self)
+        return False
+
+    def __lt__(self, other: Any) -> bool:
+        if isinstance(other, BomTarget):
+            return self.ref < other.ref
+        return NotImplemented
+
+    def __hash__(self) -> int:
+        return hash((self.ref, tuple(self.versions)))
+
+    def __repr__(self) -> str:
+        return f'<BomTarget ref={self.ref}>'
 
 
 class VulnerabilityAnalysis:
@@ -155,9 +196,9 @@ class VulnerabilityAnalysis:
         See the CycloneDX schema: https://cyclonedx.org/docs/1.4/#type_vulnerabilityType
     """
 
-    def __init__(self, state: Optional[ImpactAnalysisState] = None,
+    def __init__(self, *, state: Optional[ImpactAnalysisState] = None,
                  justification: Optional[ImpactAnalysisJustification] = None,
-                 responses: Optional[List[ImpactAnalysisResponse]] = None,
+                 responses: Optional[Iterable[ImpactAnalysisResponse]] = None,
                  detail: Optional[str] = None) -> None:
         if not state and not justification and not responses and not detail:
             raise NoPropertiesProvidedException(
@@ -166,13 +207,16 @@ class VulnerabilityAnalysis:
             )
         self.state = state
         self.justification = justification
-        self.response = responses
+        self.response = SortedSet(responses or [])
         self.detail = detail
 
     @property
     def state(self) -> Optional[ImpactAnalysisState]:
         """
         The declared current state of an occurrence of a vulnerability, after automated or manual analysis.
+
+        Returns:
+            `ImpactAnalysisState` if set else `None`
         """
         return self._state
 
@@ -184,6 +228,9 @@ class VulnerabilityAnalysis:
     def justification(self) -> Optional[ImpactAnalysisJustification]:
         """
         The rationale of why the impact analysis state was asserted.
+
+        Returns:
+            `ImpactAnalysisJustification` if set else `None`
         """
         return self._justification
 
@@ -192,17 +239,20 @@ class VulnerabilityAnalysis:
         self._justification = justification
 
     @property
-    def response(self) -> Optional[List[ImpactAnalysisResponse]]:
+    def response(self) -> "SortedSet[ImpactAnalysisResponse]":
         """
         A list of responses to the vulnerability by the manufacturer, supplier, or project responsible for the
         affected component or service. More than one response is allowed. Responses are strongly encouraged for
         vulnerabilities where the analysis state is exploitable.
+
+        Returns:
+            Set of `ImpactAnalysisResponse`
         """
         return self._response
 
     @response.setter
-    def response(self, responses: Optional[List[ImpactAnalysisResponse]]) -> None:
-        self._response = responses
+    def response(self, responses: Iterable[ImpactAnalysisResponse]) -> None:
+        self._response = SortedSet(responses)
 
     @property
     def detail(self) -> Optional[str]:
@@ -210,7 +260,9 @@ class VulnerabilityAnalysis:
         A detailed description of the impact including methods used during assessment. If a vulnerability is not
         exploitable, this field should include specific details on why the component or service is not impacted by this
         vulnerability.
-        :return:
+
+        Returns:
+            `str` if set else `None`
         """
         return self._detail
 
@@ -218,6 +270,17 @@ class VulnerabilityAnalysis:
     def detail(self, detail: Optional[str]) -> None:
         self._detail = detail
 
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, VulnerabilityAnalysis):
+            return hash(other) == hash(self)
+        return False
+
+    def __hash__(self) -> int:
+        return hash((self.state, self.justification, tuple(self.response), self.detail))
+
+    def __repr__(self) -> str:
+        return f'<VulnerabilityAnalysis state={self.state}, justification={self.justification}>'
+
 
 class VulnerabilityAdvisory:
     """
@@ -227,7 +290,7 @@ class VulnerabilityAdvisory:
         See the CycloneDX schema: https://cyclonedx.org/docs/1.4/#type_advisoryType
     """
 
-    def __init__(self, url: XsUri, title: Optional[str] = None) -> None:
+    def __init__(self, *, url: XsUri, title: Optional[str] = None) -> None:
         self.title = title
         self.url = url
 
@@ -253,6 +316,22 @@ class VulnerabilityAdvisory:
     def url(self, url: XsUri) -> None:
         self._url = url
 
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, VulnerabilityAdvisory):
+            return hash(other) == hash(self)
+        return False
+
+    def __lt__(self, other: Any) -> bool:
+        if isinstance(other, VulnerabilityAdvisory):
+            return ComparableTuple((self.title, self.url)) < ComparableTuple((other.title, other.url))
+        return NotImplemented
+
+    def __hash__(self) -> int:
+        return hash((self.title, self.url))
+
+    def __repr__(self) -> str:
+        return f'<VulnerabilityAdvisory url={self.url}, title={self.title}>'
+
 
 class VulnerabilitySource:
     """
@@ -264,7 +343,7 @@ class VulnerabilitySource:
         See the CycloneDX schema: https://cyclonedx.org/docs/1.4/#type_vulnerabilitySourceType
     """
 
-    def __init__(self, name: Optional[str] = None, url: Optional[XsUri] = None) -> None:
+    def __init__(self, *, name: Optional[str] = None, url: Optional[XsUri] = None) -> None:
         if not name and not url:
             raise NoPropertiesProvidedException(
                 'Either name or url must be provided for a VulnerabilitySource - neither provided'
@@ -294,6 +373,22 @@ class VulnerabilitySource:
     def url(self, url: Optional[XsUri]) -> None:
         self._url = url
 
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, VulnerabilitySource):
+            return hash(other) == hash(self)
+        return False
+
+    def __lt__(self, other: Any) -> bool:
+        if isinstance(other, VulnerabilitySource):
+            return ComparableTuple((self.name, self.url)) < ComparableTuple((other.name, other.url))
+        return NotImplemented
+
+    def __hash__(self) -> int:
+        return hash((self.name, self.url))
+
+    def __repr__(self) -> str:
+        return f'<VulnerabilityAdvisory name={self.name}, url={self.url}>'
+
 
 class VulnerabilityReference:
     """
@@ -308,7 +403,7 @@ class VulnerabilityReference:
         See the CycloneDX schema: https://cyclonedx.org/docs/1.4/#type_vulnerabilityType
     """
 
-    def __init__(self, id: Optional[str] = None, source: Optional[VulnerabilitySource] = None) -> None:
+    def __init__(self, *, id: Optional[str] = None, source: Optional[VulnerabilitySource] = None) -> None:
         if not id and not source:
             raise NoPropertiesProvidedException(
                 'Either id or source must be provided for a VulnerabilityReference - neither provided'
@@ -338,8 +433,24 @@ class VulnerabilityReference:
     def source(self, source: Optional[VulnerabilitySource]) -> None:
         self._source = source
 
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, VulnerabilityReference):
+            return hash(other) == hash(self)
+        return False
 
-class VulnerabilityScoreSource(Enum):
+    def __lt__(self, other: Any) -> bool:
+        if isinstance(other, VulnerabilityReference):
+            return ComparableTuple((self.id, self.source)) < ComparableTuple((other.id, other.source))
+        return NotImplemented
+
+    def __hash__(self) -> int:
+        return hash((self.id, self.source))
+
+    def __repr__(self) -> str:
+        return f'<VulnerabilityReference id={self.id}, source={self.source}>'
+
+
+class VulnerabilityScoreSource(str, Enum):
     """
     Enum object that defines the permissible source types for a Vulnerability's score.
 
@@ -409,7 +520,7 @@ class VulnerabilityScoreSource(Enum):
         return self.value
 
 
-class VulnerabilitySeverity(Enum):
+class VulnerabilitySeverity(str, Enum):
     """
     Class that defines the permissible severities for a Vulnerability.
 
@@ -476,7 +587,7 @@ class VulnerabilityRating:
         they are redundant if you have the vector (the vector allows you to calculate the scores).
     """
 
-    def __init__(self, source: Optional[VulnerabilitySource] = None, score: Optional[Decimal] = None,
+    def __init__(self, *, source: Optional[VulnerabilitySource] = None, score: Optional[Decimal] = None,
                  severity: Optional[VulnerabilitySeverity] = None,
                  method: Optional[VulnerabilityScoreSource] = None, vector: Optional[str] = None,
                  justification: Optional[str] = None,
@@ -571,6 +682,25 @@ class VulnerabilityRating:
     def justification(self, justification: Optional[str]) -> None:
         self._justification = justification
 
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, VulnerabilityRating):
+            return hash(other) == hash(self)
+        return False
+
+    def __lt__(self, other: Any) -> bool:
+        if isinstance(other, VulnerabilityRating):
+            return ComparableTuple(
+                (self.severity, self.score, self.source, self.method, self.vector, self.justification)) < \
+                ComparableTuple(
+                    (other.severity, other.score, other.source, other.method, other.vector, other.justification))
+        return NotImplemented
+
+    def __hash__(self) -> int:
+        return hash((self.source, self.score, self.severity, self.method, self.vector, self.justification))
+
+    def __repr__(self) -> str:
+        return f'<VulnerabilityRating score={self.score}, source={self.source}>'
+
 
 class VulnerabilityCredits:
     """
@@ -583,42 +713,53 @@ class VulnerabilityCredits:
         See the CycloneDX schema: https://cyclonedx.org/docs/1.4/#type_vulnerabilityType
     """
 
-    def __init__(self, organizations: Optional[List[OrganizationalEntity]] = None,
-                 individuals: Optional[List[OrganizationalContact]] = None) -> None:
+    def __init__(self, *, organizations: Optional[Iterable[OrganizationalEntity]] = None,
+                 individuals: Optional[Iterable[OrganizationalContact]] = None) -> None:
         if not organizations and not individuals:
             raise NoPropertiesProvidedException(
                 'One of `organizations` or `individuals` must be populated - neither were'
             )
-        self.organizations = organizations
-        self.individuals = individuals
+        self.organizations = SortedSet(organizations or [])
+        self.individuals = SortedSet(individuals or [])
 
     @property
-    def organizations(self) -> Optional[List[OrganizationalEntity]]:
+    def organizations(self) -> "SortedSet[OrganizationalEntity]":
         """
         The organizations credited with vulnerability discovery.
 
         Returns:
-             List of `OrganizationalEntity` or `None`
+             Set of `OrganizationalEntity`
         """
         return self._organizations
 
     @organizations.setter
-    def organizations(self, organizations: Optional[List[OrganizationalEntity]]) -> None:
-        self._organizations = organizations
+    def organizations(self, organizations: Iterable[OrganizationalEntity]) -> None:
+        self._organizations = SortedSet(organizations)
 
     @property
-    def individuals(self) -> Optional[List[OrganizationalContact]]:
+    def individuals(self) -> "SortedSet[OrganizationalContact]":
         """
         The individuals, not associated with organizations, that are credited with vulnerability discovery.
 
         Returns:
-            List of `OrganizationalContact` or `None`
+            Set of `OrganizationalContact`
         """
         return self._individuals
 
     @individuals.setter
-    def individuals(self, individuals: Optional[List[OrganizationalContact]]) -> None:
-        self._individuals = individuals
+    def individuals(self, individuals: Iterable[OrganizationalContact]) -> None:
+        self._individuals = SortedSet(individuals)
+
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, VulnerabilityCredits):
+            return hash(other) == hash(self)
+        return False
+
+    def __hash__(self) -> int:
+        return hash((tuple(self.organizations), tuple(self.individuals)))
+
+    def __repr__(self) -> str:
+        return f'<VulnerabilityCredits id={id(self)}>'
 
 
 class Vulnerability:
@@ -632,36 +773,36 @@ class Vulnerability:
         See the CycloneDX schema: https://cyclonedx.org/docs/1.4/#type_vulnerabilityType
     """
 
-    def __init__(self, bom_ref: Optional[str] = None, id: Optional[str] = None,
+    def __init__(self, *, bom_ref: Optional[str] = None, id: Optional[str] = None,
                  source: Optional[VulnerabilitySource] = None,
-                 references: Optional[List[VulnerabilityReference]] = None,
-                 ratings: Optional[List[VulnerabilityRating]] = None, cwes: Optional[List[int]] = None,
+                 references: Optional[Iterable[VulnerabilityReference]] = None,
+                 ratings: Optional[Iterable[VulnerabilityRating]] = None, cwes: Optional[Iterable[int]] = None,
                  description: Optional[str] = None, detail: Optional[str] = None, recommendation: Optional[str] = None,
-                 advisories: Optional[List[VulnerabilityAdvisory]] = None, created: Optional[datetime] = None,
+                 advisories: Optional[Iterable[VulnerabilityAdvisory]] = None, created: Optional[datetime] = None,
                  published: Optional[datetime] = None, updated: Optional[datetime] = None,
                  credits: Optional[VulnerabilityCredits] = None,
-                 tools: Optional[List[Tool]] = None, analysis: Optional[VulnerabilityAnalysis] = None,
-                 affects_targets: Optional[List[BomTarget]] = None,
+                 tools: Optional[Iterable[Tool]] = None, analysis: Optional[VulnerabilityAnalysis] = None,
+                 affects_targets: Optional[Iterable[BomTarget]] = None,
                  # Deprecated Parameters kept for backwards compatibility
                  source_name: Optional[str] = None, source_url: Optional[str] = None,
-                 recommendations: Optional[List[str]] = None) -> None:
-        self.bom_ref = bom_ref or str(uuid4())
+                 recommendations: Optional[Iterable[str]] = None) -> None:
+        self._bom_ref = BomRef(value=bom_ref)
         self.id = id
         self.source = source
-        self.references = references or []
-        self.ratings = ratings or []
-        self.cwes = cwes or []
+        self.references = SortedSet(references or [])
+        self.ratings = SortedSet(ratings or [])
+        self.cwes = SortedSet(cwes or [])
         self.description = description
         self.detail = detail
         self.recommendation = recommendation
-        self.advisories = advisories or []
+        self.advisories = SortedSet(advisories or [])
         self.created = created
         self.published = published
         self.updated = updated
         self.credits = credits
-        self.tools = tools or []
+        self.tools = SortedSet(tools or [])
         self.analysis = analysis
-        self.affects = affects_targets or []
+        self.affects = SortedSet(affects_targets or [])
 
         if source_name or source_url:
             warnings.warn('`source_name` and `source_url` are deprecated - use `source`', DeprecationWarning)
@@ -671,28 +812,27 @@ class Vulnerability:
         if recommendations:
             warnings.warn('`recommendations` is deprecated - use `recommendation`', DeprecationWarning)
             if not recommendation:
-                self.recommendation = recommendations.pop()
+                self.recommendation = next(iter(recommendations))
 
     @property
-    def bom_ref(self) -> Optional[str]:
+    def bom_ref(self) -> BomRef:
         """
         Get the unique reference for this Vulnerability in this BOM.
 
         If a value was not provided in the constructor, a UUIDv4 will have been assigned.
 
         Returns:
-           `str` unique identifier for this Vulnerability
+           `BomRef`
         """
         return self._bom_ref
 
-    @bom_ref.setter
-    def bom_ref(self, bom_ref: Optional[str]) -> None:
-        self._bom_ref = bom_ref
-
     @property
     def id(self) -> Optional[str]:
         """
         The identifier that uniquely identifies the vulnerability. For example: CVE-2021-39182.
+
+        Returns:
+            `str` if set else `None`
         """
         return self._id
 
@@ -704,6 +844,9 @@ class Vulnerability:
     def source(self) -> Optional[VulnerabilitySource]:
         """
         The source that published the vulnerability.
+
+        Returns:
+            `VulnerabilitySource` if set else `None`
         """
         return self._source
 
@@ -712,83 +855,60 @@ class Vulnerability:
         self._source = source
 
     @property
-    def references(self) -> List[VulnerabilityReference]:
+    def references(self) -> "SortedSet[VulnerabilityReference]":
         """
         Zero or more pointers to vulnerabilities that are the equivalent of the vulnerability specified. Often times,
         the same vulnerability may exist in multiple sources of vulnerability intelligence, but have different
         identifiers. References provides a way to correlate vulnerabilities across multiple sources of vulnerability
         intelligence.
+
+        Returns:
+            Set of `VulnerabilityReference`
         """
         return self._references
 
     @references.setter
-    def references(self, references: List[VulnerabilityReference]) -> None:
-        self._references = references
-
-    def add_reference(self, reference: VulnerabilityReference) -> None:
-        """
-        Add an additional reference for this Vulnerability.
-
-        Vulnerabilities may benefit from pointers to vulnerabilities that are the equivalent of the vulnerability
-        specified. Often times, the same vulnerability may exist in multiple sources of vulnerability intelligence, but
-        have different identifiers. These references provide a way to correlate vulnerabilities across multiple sources
-        of vulnerability intelligence.
-
-        Args:
-            reference:
-                `VulnerabilityReference` reference to add
-        """
-        self.references = self.references + [reference]
+    def references(self, references: Iterable[VulnerabilityReference]) -> None:
+        self._references = SortedSet(references)
 
     @property
-    def ratings(self) -> List[VulnerabilityRating]:
+    def ratings(self) -> "SortedSet[VulnerabilityRating]":
         """
         List of vulnerability ratings.
+
+        Returns:
+            Set of `VulnerabilityRating`
         """
         return self._ratings
 
     @ratings.setter
-    def ratings(self, ratings: List[VulnerabilityRating]) -> None:
-        self._ratings = ratings
-
-    def add_rating(self, rating: VulnerabilityRating) -> None:
-        """
-        Add a vulnerability rating.
-
-        Args:
-            rating:
-                `VulnerabilityRating`
-        """
-        self.ratings = self.ratings + [rating]
+    def ratings(self, ratings: Iterable[VulnerabilityRating]) -> None:
+        self._ratings = SortedSet(ratings)
 
     @property
-    def cwes(self) -> List[int]:
+    def cwes(self) -> "SortedSet[int]":
         """
         A list of CWE (Common Weakness Enumeration) identifiers.
 
         .. note::
             See https://cwe.mitre.org/
+
+        Returns:
+            Set of `int`
         """
         return self._cwes
 
     @cwes.setter
-    def cwes(self, cwes: List[int]) -> None:
-        self._cwes = cwes
-
-    def add_cwe(self, cwe: int) -> None:
-        """
-        Add a CWE identifier.
-
-        Args:
-            cwe:
-                `int` identifier for the CWE
-        """
-        self.cwes = self.cwes + [cwe]
+    def cwes(self, cwes: Iterable[int]) -> None:
+        self._cwes = SortedSet(cwes)
 
     @property
     def description(self) -> Optional[str]:
         """
         A description of the vulnerability as provided by the source.
+
+        Returns:
+            `str` if set else `None`
         """
         return self._description
 
@@ -801,6 +921,9 @@ class Vulnerability:
         """
         If available, an in-depth description of the vulnerability as provided by the source organization. Details
         often include examples, proof-of-concepts, and other information useful in understanding root cause.
+
+        Returns:
+            `str` if set else `None`
         """
         return self._detail
 
@@ -812,6 +935,9 @@ class Vulnerability:
     def recommendation(self) -> Optional[str]:
         """
         Recommendations of how the vulnerability can be remediated or mitigated.
+
+        Returns:
+            `str` if set else `None`
         """
         return self._recommendation
 
@@ -820,28 +946,27 @@ class Vulnerability:
         self._recommendation = recommendation
 
     @property
-    def advisories(self) -> List[VulnerabilityAdvisory]:
+    def advisories(self) -> "SortedSet[VulnerabilityAdvisory]":
         """
         Advisories relating to the Vulnerability.
+
+        Returns:
+            Set of `VulnerabilityAdvisory`
         """
         return self._advisories
 
     @advisories.setter
-    def advisories(self, advisories: List[VulnerabilityAdvisory]) -> None:
-        self._advisories = advisories
-
-    def add_advisory(self, advisory: VulnerabilityAdvisory) -> None:
-        """
-        Add a advisory.
-
-        Args:
-            advisory:
-                `VulnerabilityAdvisory`
-        """
-        self.advisories = self.advisories + [advisory]
+    def advisories(self, advisories: Iterable[VulnerabilityAdvisory]) -> None:
+        self._advisories = SortedSet(advisories)
 
     @property
     def created(self) -> Optional[datetime]:
+        """
+        The date and time (timestamp) when the vulnerability record was created in the vulnerability database.
+
+        Returns:
+            `datetime` if set else `None`
+        """
         return self._created
 
     @created.setter
@@ -850,6 +975,12 @@ class Vulnerability:
 
     @property
     def published(self) -> Optional[datetime]:
+        """
+        The date and time (timestamp) when the vulnerability record was first published.
+
+        Returns:
+            `datetime` if set else `None`
+        """
         return self._published
 
     @published.setter
@@ -858,6 +989,12 @@ class Vulnerability:
 
     @property
     def updated(self) -> Optional[datetime]:
+        """
+        The date and time (timestamp) when the vulnerability record was last updated.
+
+        Returns:
+            `datetime` if set else `None`
+        """
         return self._updated
 
     @updated.setter
@@ -868,6 +1005,9 @@ class Vulnerability:
     def credits(self) -> Optional[VulnerabilityCredits]:
         """
         Individuals or organizations credited with the discovery of the vulnerability.
+
+        Returns:
+            `VulnerabilityCredits` if set else `None`
         """
         return self._credits
 
@@ -876,30 +1016,26 @@ class Vulnerability:
         self._credits = credits
 
     @property
-    def tools(self) -> List[Tool]:
+    def tools(self) -> "SortedSet[Tool]":
         """
         The tool(s) used to identify, confirm, or score the vulnerability.
+
+        Returns:
+            Set of `Tool`
         """
         return self._tools
 
     @tools.setter
-    def tools(self, tools: List[Tool]) -> None:
-        self._tools = tools
-
-    def add_tool(self, tool: Tool) -> None:
-        """
-        Add a tool used to identify, confirm, or score the vulnerability.
-
-        Args:
-            tool:
-                `Tool`
-        """
-        self.tools = self.tools + [tool]
+    def tools(self, tools: Iterable[Tool]) -> None:
+        self._tools = SortedSet(tools)
 
     @property
     def analysis(self) -> Optional[VulnerabilityAnalysis]:
         """
         Analysis of the Vulnerability in your context.
+
+        Returns:
+            `VulnerabilityAnalysis` if set else `None`
         """
         return self._analysis
 
@@ -908,70 +1044,38 @@ class Vulnerability:
         self._analysis = analysis
 
     @property
-    def affects(self) -> Optional[List[BomTarget]]:
+    def affects(self) -> "SortedSet[BomTarget]":
         """
         The components or services that are affected by the vulnerability.
+
+        Returns:
+            Set of `BomTarget`
         """
         return self._affects
 
     @affects.setter
-    def affects(self, affects_targets: Optional[List[BomTarget]]) -> None:
-        self._affects = affects_targets
+    def affects(self, affects_targets: Iterable[BomTarget]) -> None:
+        self._affects = SortedSet(affects_targets)
 
-    # Methods pre-dating 1.4 that are kept for some backwards compatability - they will be removed in a future release!
-    def get_source_name(self) -> Optional[str]:
-        """
-        Prior to Schema Version 1.4 when Vulnerabilities were supported by a schema extension, `source_name` and
-        `source_url` where represented differently in the model.
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, Vulnerability):
+            return hash(other) == hash(self)
+        return False
 
-        .. warning::
-            Deprecated - this method will be removed in a future version.
+    def __lt__(self, other: Any) -> bool:
+        if isinstance(other, Vulnerability):
+            return ComparableTuple(
+                (self.id, self.description, self.detail, self.source, self.created, self.published)) < \
+                ComparableTuple(
+                    (other.id, other.description, other.detail, other.source, other.created, other.published))
+        return NotImplemented
 
-            See `Vulnerability.source.get_name()`
-        """
-        warnings.warn(
-            'The source of a Vulnerability is represnted differently in Schema Versions >= 1.4. '
-            'Vulnerability.get_source_name() is deprecated and will be removed in a future release. '
-            'Use Vulnerability.source.get_name()',
-            DeprecationWarning
-        )
-        if self.source:
-            return self.source.name
-        return None
+    def __hash__(self) -> int:
+        return hash((
+            self.id, self.source, tuple(self.references), tuple(self.ratings), tuple(self.cwes), self.description,
+            self.detail, self.recommendation, tuple(self.advisories), self.created, self.published, self.updated,
+            self.credits, tuple(self.tools), self.analysis, tuple(self.affects)
+        ))
 
-    def get_source_url(self) -> Optional[ParseResult]:
-        """
-        Prior to Schema Version 1.4 when Vulnerabilities were supported by a schema extension, `source_name` and
-        `source_url` where represented differently in the model.
-
-        .. warning::
-            Deprecated - this method will be removed in a future version.
-
-            See `Vulnerability.source.get_url()`
-        """
-        warnings.warn(
-            'The source of a Vulnerability is represnted differently in Schema Versions >= 1.4. '
-            'Vulnerability.get_source_name() is deprecated and will be removed in a future release. '
-            'Use Vulnerability.source.get_url()',
-            DeprecationWarning
-        )
-        if self.source and self.source.url:
-            return urlparse(str(self.source.url))
-        return None
-
-    def get_recommendations(self) -> List[str]:
-        """
-        Prior to Schema Version 1.4 when Vulnerabilities were supported by a schema extension, multiple recommendations
-        where permissible.
-
-        .. warning::
-            Deprecated - this method will be removed in a future version.
-
-            See `Vulnerability.recommendation`
-        """
-        warnings.warn(
-            'A Vulnerability has only a single recommendation from Schema Version >= 1.4. '
-            'Vulnerability.get_recommendations() is deprecated and will be removed in a future release.',
-            DeprecationWarning
-        )
-        return [self.recommendation] if self.recommendation else []
+    def __repr__(self) -> str:
+        return f'<Vulnerability bom-ref={self.bom_ref}, id={self.id}>'
